'use strict';

const debug = require('debug')('next-cli');
const EventEmitter = require('events');
const parser = require('yargs-parser');
const co = require('co');
const colors = require('colors/safe');


/**
 *
 */
class Command extends EventEmitter {

  constructor(parent) {
    super();

    this._parent = parent || null;

    this._name = '';
    this._alias = '';

    this._version = '';
    this._description = '';
    this._example = '';

    this._group = '';
    this._hide = false;
    this._allowUnknownOption = false;

    this._commands = [];
    this._args = [];
    this._opts = [];

    this._actionFn = null;
  }

  name(n) {
    if (n) {
      this._name = n;
      return this;
    }
    return this._name;
  }

  alias(a) {
    if (a) {
      if (this._parent) {
        let cmd = this._parent.find(a);
        if (cmd) {
          throw new Error('the alias `' + a + '` already define by `' + cmd.name() + '(' + cmd.alias() + ')`.');
        }
        this._alias = a;
      }
      return this;
    }
    return this._alias;
  }

  version(v) {
    if (v) {
      this._version = v;
      return this;
    }
    return this._version;
  }

  description(desc) {
    if (desc) {
      this._description = desc;
      return this;
    }
    return this._description;
  }

  example(str) {
    if (str) {
      this._example = str;
      return this;
    }
    return this._example;
  }

  command(spec, desc) {
    if (!spec) {
      return this;
    }
    if (spec == this._name) {
      return this;
    }

    const parents = this.parents();
    if (parents.length > 2) {
      throw new Error('the max sub commands level only allow 3.');
    }

    spec = spec.split(/ +/);

    const cmd = new Command(this);
    cmd.name(spec.shift());
    cmd.description(desc);
    spec.forEach(arg => cmd.argument(arg));

    // avoid duplicate
    if (this._commands.filter(c => c._name == cmd._name).length == 0) {
      this._commands.push(cmd);
    }

    return cmd;
  }

  argument(flag, desc, choices) {
    let arg = {
      'name': '',
      'desc': '',
      'variadic': false,
      'required': false,
      'choices': null
    };

    switch (flag[0]) {
      case '<':
        arg.required = true;
        arg.name = flag.slice(1, -1);
        break;
      case '[':
        arg.name = flag.slice(1, -1);
        break;
    }
    if (arg.name.length > 3 && arg.name.slice(-3) === '...') {
      arg.variadic = true;
      arg.name = arg.name.slice(0, - 3);
    }

    if (!arg.name) {
      throw new Error('the arg `' + flag + '` missed name.');
    }

    if (!arg.variadic && choices) {
      arg.choices = choices;
    }

    if (desc) {
      arg.desc = desc;
    }

    let idx = -1;
    for (let i = 0, len = this._args.length; i < len; i += 1) {
      if (this._args[i].name == arg.name) {
        idx = i;
        break;
      }
    }

    if (idx != -1) {
      this._args[idx] = arg;
    } else {
      this._args.push(arg);
    }

    return this;
  }

  option(flag, desc, conf, defVal) {
    let config = {};

    let len = arguments.length;
    if (len > 2) {
      if (len == 3) {
        // conf->default or config
        if (null != conf) {
          if ('[object Object]' == conf.toString()) {
            config = conf;
          } else {
            config['default'] = conf;
          }
        } else {
          config = conf || {};
        }
      }
      if (len == 4) {
        if (typeof conf == 'function') {
          config['parse'] = conf;
        }
        if (null != defVal && '[object Object]' != defVal.toString()) {
          config['default'] = defVal;
        }
      }
    }

    var opt = {
      'name': '',
      'alias': '',
      'required': false,
      'flag': '',
      'desc': '',
      'type': '',
      'defaultVal': null,
      'choices': null,
      'hide': false,
      'pass': true,
      'narg': null,
      'parse': null
    };

    opt.flag = flag;
    opt.desc = desc || '';

    flag = flag.split(/[ ,|]+/);
    if (flag.length > 1 && !/^[[<]/.test(flag[1])) {
      opt.alias = flag.shift().replace(/^(--|-)/, '');
    }
    opt.name = flag.shift().replace(/^(--|-)/, '');

    if (this._opts.filter(o => o.name == opt.name && o.alias == opt.alias).length > 1) {
      throw new Error('the opt `' + opt.flag + '` already defined.');
    }

//    opt.type = 'boolean';
    if (flag[0]) {
      if (/^[[<]/.test(flag[0])) {
        opt.narg = 1;

        opt.defaultVal = config['default'];

        if (Array.isArray(config['choices'])) {
          opt.choices = config['choices'];
        }

        if (/^</.test(flag[0])) {
          opt.required = true;
        }

        opt.type = 'string';
        if (typeof opt.defaultVal == 'number') {
          opt.type = 'number';
        }
        if (flag[0].indexOf('...') != -1) {
          opt.type = 'array';
          opt.narg = null;
          opt.defaultVal = [];
        }
      }
    }

    if (config['hide'] != null) {
      opt.hide = config['hide'];
    }
    if (config['pass'] != null) {
      opt.pass = config['pass'];
    }
    if (typeof config['parse'] == 'function') {
      opt.parse = config['parse'];
    }

    this._opts.push(opt);

    return this;
  }

  usage(str) {
    if (str) {
      str = str.split(/ +/);
      str.forEach(s => this.argument(s));
      return this;
    }
    return this._name + ' ' + this._argument();
  }

  group(name) {
    if (name) {
      this._group = name;
      return this;
    }
    return this._group;
  }

  hide() {
    this._hide = true;
    return this;
  }

  action(fn) {
    if (typeof fn == 'function') {
      this._actionFn = fn;
      return this;
    }
    return this._actionFn;
  }

  end() {
    return this._parent;
  }

  find(arg) {
    for (let i = 0, len = this._commands.length; i < len; i += 1) {
      let cmd = this._commands[i];
      if (cmd.is(arg)) {
        return cmd;
      }
    }
    return null;
  }

  is(arg) {
    return this._name == arg || this._alias == arg;
  }

  parents() {
    const arr = [];

    let p = this._parent;
    while (p) {
      arr.push(p);
      p = p._parent;
    }

    return arr;
  }

  allowUnknownOption() {
    this._allowUnknownOption = true;
    return this;
  }

  help() {
    let usageStr = this._usageHelp();
    let descStr = this.description() || '';
    let argumentStr = this._argumentHelp();
    let commandStr = this._commandHelp();
    let optionStr = this._optionHelp();

    let str = [];
    if (usageStr) {
      str.push(usageStr);
    }
    if (descStr) {
      str.push(descStr);
    }
    if (argumentStr) {
      str.push(argumentStr);
    }
    if (commandStr) {
      str.push(commandStr);
    }
    if (optionStr) {
      str.push(optionStr);
    }

    return '\n' + str.join('\n\n');
  }

  showHelp() {
    console.log(this.help());
  }

  * parse(argv, config) {
    config = config || {};

    if (this._parent != null) {
      throw new Error('`parse` should called at root command.');
    }

    argv = argv.slice();

    let rawArgv = argv.slice();
    let remainedArgv = [];
    let optsConfig, parsed;
    let idx;

    // process --
    idx = argv.indexOf('--');
    if (idx != -1) {
      remainedArgv = argv.slice(idx + 1);
      argv = argv.slice(0, idx);
    }

    optsConfig = {
      'alias': {},
      'default': {},
      'array': [],
      'boolean': [],
      'count': [],
      'string': [],
      'number': [],
      'narg': {},
      'configuration': {
        'boolean-negation': false,
        'camel-case-expansion': false
      }
    };

    parsed = {
      '--': remainedArgv,
      'raw': rawArgv,
      'cmd': null,
      'argv': null,
      'args': null,
      'opts': null
    };

    this._parse(argv, optsConfig, parsed);
    this.emit('parsed', {'parsed': parsed});

    if (config.use == 'parser') {
      return parsed;
    }

    if (!parsed.cmd) {
      return {'cmd': parsed.argv._[0], '$error': 'invalid command'};
    }

    if (parsed.argv.version) {
      let v = parsed.cmd.version() || this.version();
      console.log(v);
      return v;
    }
    if (parsed.argv.help) {
      let help = parsed.cmd.help();
      console.log(help);
      return help;
    }

    let errArg = parsed.args.filter(arg => arg.err)[0];
    if (errArg) {
      parsed.cmd.error(errArg.err + ' `' + errArg.name + '`.', true);
      return {'$error': 'invalid arg'};
    }

    let errOpt = null;
    for (let k in parsed.opts.opts) {
      if (parsed.opts.opts[k].err) {
        errOpt = parsed.opts.opts[k];
        break;
      }
    }
    if (errOpt) {
      parsed.cmd.error(errOpt.err + ' `' + errOpt.opt.flag + '`.', true);
      return {'$error': 'invalid opt'};
    }

    if (!parsed.cmd._allowUnknownOption) {
      let unknowOpts = Object.keys(parsed.opts.unknown);
      if (unknowOpts.length) {
        parsed.cmd.error(Command.lang.unknownOption + ': ' + unknowOpts.map(o => '`' + o + '`').join(', ') + '.', true);
        return {'$error': 'unknown opt'};
      }
    }

    let actionFn = parsed.cmd.action();

    if (typeof actionFn != 'function') {
      this.error('can not find action fn of command `' + parsed.cmd.name() + '`.');
      return {'$error': 'invalid action fn'};
    }

    let applyArgs, applyOpts, applyCtx;

    let applyData = this._apply(parsed);

    applyArgs = applyData.args;
    applyOpts = applyData.opts;
    applyCtx = applyData.ctx;

    // apply ctx.
    if (typeof config.applyCtx == 'function') {
      applyCtx = config.applyCtx(applyCtx);
    }

    this.emit('command.start', {
      'cmd': parsed.cmd,
      'args': applyArgs
    });
    return yield co.apply(null, [actionFn.bind(applyCtx)].concat(applyArgs));
  }

  error(err, showHelp) {
    let str = colors.red(Command.lang.error + err);
    if (showHelp) {
      str += ' ' + Command.lang.showHelp;
    }
    console.error();
    console.error(str);

    if (showHelp) {
      this.showHelp();
    }
  }

  _apply(parsed) {
    let applyArgs = [];
    let applyOpts = {};
    let applyCtx = {'cmd': parsed.cmd};

    // apply args.
    parsed.args.forEach(arg => applyArgs.push(arg.val));

    // apply opts.
    for (let k in parsed.opts.opts) {
      applyOpts[camelcase(k)] = parsed.opts.opts[k].val;
    }
    applyOpts['--'] = parsed['--'];
    applyOpts['$'] = parsed.opts.unknown;
    applyArgs.push(applyOpts);

    return {
      'args': applyArgs,
      'opts': applyOpts,
      'ctx': applyCtx
    };
  }

  _parse(argv, optsConfig, parsed) {
    optsConfig = this._optsConfig(optsConfig);

    parsed.argv = parser(argv, optsConfig);
    parsed.args = this._parseArgs(parsed.argv._);
    parsed.opts = this._parseOpts(parsed.argv, parsed.opts);

    let ctx = parsed.cmd || this;
    let name = parsed.argv._[0];
    let cmd = ctx.find(name);
    if (cmd) {
      let idx = argv.indexOf(name);
      if (idx != -1) {
        argv.splice(idx, 1);
      }
      parsed.cmd = cmd;

      return cmd._parse(argv, optsConfig, parsed);
    }

    if (!name) {
      parsed.cmd = this;
    }
  }

  _parseArgs(argv) {
    const args = [];

    let i = 0;
    argv = argv.slice();
    this._args.forEach((arg) => {
      if (arg.required && null == argv[i]) {
        args.push({
          'arg': arg,
          'name': arg.name,
          'val': null,
          'err': Command.lang.missingRequiredArgument
        });
        i += 1;
      } else if (arg.variadic) {
        args.push({
          'arg': arg,
          'name': arg.name,
          'val': argv.splice(i)
        });
        i += 1;
      } else {
        if (null != argv[i] && arg.choices) {
          let choices = Array.isArray(arg.choices) ? arg.choices : Object.keys(arg.choices);
          if (choices.indexOf(argv[i]) == -1) {
            if (choices.indexOf('*') == -1) {
              args.push({
                'arg': arg,
                'name': arg.name,
                'val': null,
                'err': Command.lang.invalidArgument
              });
              i += 1;
            } else {
              args.push({
                'arg': arg,
                'name': arg.name,
                'val': ''
              });
            }
          } else {
            args.push({
              'arg': arg,
              'name': arg.name,
              'val': argv[i]
            });
            i += 1;
          }
        } else {
          args.push({
            'arg': arg,
            'name': arg.name,
            'val': argv[i]
          });
          i += 1;
        }
      }
    });

    return args;
  }

  _parseOpts(argv, opts) {
    opts = opts || {
      'opts': {},
      'unknown': {}
    };

    const passedOpts = Object.assign({}, argv);
    this._mergedOpts().forEach(opt => {
      if (opt.name in passedOpts) {
        let val = passedOpts[opt.name];
        if (val == null || typeof val == 'number' && Number.isNaN(val)) {
          if (opt.defaultVal != null) {
            val = opt.defaultVal;
          }
        }

        opts.opts[opt.name] = {
          'opt': opt,
          'val': val
        };

        if (opt.required && null == val) {
//          opts.opts[opt.name].err = 'missing required option argument';
        } else {
          if (Array.isArray(opt.choices) && -1 == opt.choices.indexOf(val)) {
            opts.opts[opt.name].err = Command.lang.invalidOptionArgument;
          }
        }

        if (typeof opt.parse == 'function') {
          opts.opts[opt.name].val = opt.parse(opts.opts[opt.name].val);
        }

        delete passedOpts[opt.name];
        if (opt.alias in passedOpts) {
          delete passedOpts[opt.alias];
        }
      } else {
//        if (opt.required) {
//          opts.opts[opt.name] = {
//            'opt': opt,
//            'val': null,
//            'err': 'missing required option argument'
//          };
//        }
      }

      if (!opt.pass && (opt.name in opts.opts)) {
        delete opts.opts[opt.name];
      }
    });

    const unknown = {};
    Object.keys(passedOpts).filter(k => k != '_').forEach(k => unknown[k] = passedOpts[k]);
    opts.unknown = unknown;

    return opts;
  }

  _optsConfig(config) {
    this._opts.forEach(opt => {
      if (opt.alias) {
        config.alias[opt.name] = opt.alias;
      }
      if (opt.defaultVal != null) {
        config.default[opt.name] = opt.defaultVal;
      }
      if (opt.narg != null) {
        config.narg[opt.name] = opt.narg;
      }

      if (config[opt.type] && (config[opt.type].indexOf(opt.name) == -1)) {
        config[opt.type].push(opt.name);
      }
    });

    return config;
  }

  _mergedOpts() {
    let opts = this._opts.slice();

    let p = this._parent;
    while (p) {
      opts = opts.concat(p._opts.slice());
      p = p._parent;
    }

    return opts;
  }

  _usageHelp() {
    let usage = [];

    let prefix = this._name;
    if (this._alias) {
      prefix += ' | ' + this._alias;
    }
    let p = this._parent;
    while (p) {
      prefix = p._name + ' ' + prefix;
      p = p._parent;
    }
    prefix = prefix.trim();

    if (this._commands.length > 0) {
      usage.push(prefix, '[command]', '[--options]');
      if (this._args.length > 0) {
        usage.push(' || ');
        addArgs.call(this, usage, prefix);
      }
    } else {
      addArgs.call(this, usage, prefix);
    }

    return Command.lang.usage + usage.join(' ');

    function addArgs(usage, prefix) {
      usage.push(prefix);
      if (this._args.length > 0) {
        usage.push(this._argument());
      }
      usage.push('[--options]');
    }
  }

  _argumentHelp() {
    let argument = [];

    let max = this._args.reduce((max, arg) => {
      if (arg.desc) {
        max = Math.max(max, arg.name.length);
      }
      return max;
    }, 0);

    let str;
    this._args.forEach(arg => {
      if (arg.desc) {
        str = '  ' + pad(arg.name, max) + '  ' + arg.desc + getChoices(arg);
        argument.push(str);
      }
    });

    if (argument.length > 0) {
      argument.unshift(Command.lang.arguments);
    }

    return argument.join('\n');

    function getChoices(arg) {
      if (!arg.choices) {
        return '';
      }

      let a = '';
      if (Array.isArray(arg.choices) && arg.choices.length) {
        a += ' (' + arg.choices.join(' | ') + ')';
      } else {
        let m = Object.keys(arg.choices).reduce((max, c) => Math.max(max, c.length), 0);
        for (let k in arg.choices) {
          a += '\n' + '    > ' + pad(k, m) + '  ' + arg.choices[k];
        }
      }
      return a;
    }
  }

  _optionHelp() {
    let opts = [];

    let max = this._mergedOpts()
      .filter(o => !o.hide)
      .reduce((max, o) => Math.max(max, o.flag.length), 0);

    let _opts;
    _opts = this._opts.filter(opt => !opt.hide);
    if (_opts.length > 0) {
      _opts.forEach(opt => opts.push(getOpt(opt)));
    }

    let p = this._parent;
    while (p) {
      _opts = p._opts.filter(opt => !opt.hide);
      if (_opts.length) {
        if (opts.length > 0) {
          opts.push('');
        }
        _opts.forEach(opt => opts.push(getOpt(opt)));
      }
      p = p._parent;
    }

    if (opts.length > 0) {
      opts.unshift(Command.lang.options);
    }

    return opts.join('\n');

    function getOpt(opt) {
      let str = '';
      str += '  ' + pad(opt.flag, max) + '  ' + opt.desc;
      if (Array.isArray(opt.choices) && opt.choices.length) {
        str += ' (' + opt.choices.join(' | ') + ')';
      }
      return str;
    }
  }

  _commandHelp() {
    let cmds = [];

    let groups = {};
    this._commands.forEach(c => {
      let g = c.group() || Command.lang.commands;
      if (!groups[g]) {
        groups[g] = [];
      }

      if (!c._hide) {
        groups[g].push(c);
      }
    });

    let max = this._commands.reduce((max, c) => Math.max(max, c.usage().length), 0);

    for (let k in groups) {
      groups[k].forEach(cmd => {
        cmds.push('  ' + pad(cmd.usage(), max) + '  ' + cmd.description());
      });
      cmds.push('');
    }
    cmds = cmds.slice(0, -1);

    if (cmds.length > 0) {
      cmds.unshift(Command.lang.commands);
    }

    return cmds.join('\n');
  }

  _argument() {
    return this._args
      .map(arg => {
        let nameOutput = arg.name + (arg.variadic === true ? '...' : '');
        return arg.required ? '<' + nameOutput + '>' : '[' + nameOutput + ']';
      })
      .join(' ');
  }

}

Command.lang = {
  'usage': 'Usage: ',
  'arguments': 'Arguments: ',
  'options': 'Options: ',
  'examples': 'Examples: ',
  'commands': 'Commands: ',
  'choices': 'choices: ',

  'help': 'output the help information.',
  'version': 'output the version number.',

  'error': 'ERROR: ',

  'invalidArgument': 'invalid argument',
  'missingRequiredArgument': 'missing required argument',

  'unknownOption': 'unknown option',
  'invalidOptionArgument': 'invalid option argument',

  'showHelp': 'Show Help: '
};


function Cli(config) {
  config = config || {};

  Command.lang = Object.assign({}, Command.lang, config.lang || {});

  const cmd = new Command();
  cmd
    .name(config.name)
    .description(config.description)
    .version(config.version)
    .option('-h, --help', Command.lang.help)
    .option('-v, --version', Command.lang.version)
    .action(() => cmd.showHelp());

  return cmd;
}

Cli.Command = Command;


module.exports = Cli;



function pad(str, width) {
  let len = Math.max(0, width - str.length);
  return str + Array(len + 1).join(' ');
}

function camelcase(flag) {
  return flag.split('-').reduce(function(str, word) {
    return str + word[0].toUpperCase() + word.slice(1);
  });
}