'use strict';


module.exports = function(def) {

  def.cli
    .command('install <name>', '安装 DEF 模块（插件、套件）')
    .alias('i')
    .option('-f, --force', '强制进行覆盖安装')
    .group('store')
    .action(function* (name, opts) {
      const spinner = def.ui.spinner('获取模块信息...');
      const info = yield def.store.diff(name);
      spinner.stop();

      if (info.error) {
        return def.log.error(info.error);
      }

      const locals = yield def.store.locals();
      if (locals[name] && locals[name].debug) {
        return def.log.error('[store]', '该模块处于调试模式，防止意外覆盖，禁止重新安装');
      }

      if (!info.local) {
        def.log.info('[store]', '本地不存在该模块，全新安装...');
      } else {
        if (opts.force) {
          def.log.info('[store]', '本地已存在该模块，覆盖安装...');
        } else {
          const answer = yield def.ui.confirm('本地已存在该模块，是否需要覆盖安装？');
          if (!answer) {
            return def.log.info('[store]', '取消安装');
          } else {
            opts.force = true;
          }
        }
      }

      const start = Date.now();
      def.log.info('[store]', '开始安装模块：' + info.long + '，请等待...');
      yield def.store.install(info, {'force': opts.force});
      def.log.info('[store]', '安装完成，耗时：' + (Date.now() - start) / 1000 + 's');
    });

  def.cli
    .command('update [names...]', '更新本地安装的 DEF 模块')
    .alias('up')
    .option('-C, --no-confirm', '不显示确认框')
    .group('store')
    .action(function* (names, opts) {
      let mods = [];

      if (names.length == 0) {
        def.log.info('[store]', '未指定模块，更新本地所有模块');
        mods = yield def.store.list();
      } else {
        mods = names.map(n => {
          return {'name': n};
        });
      }

      const spinner = def.ui.spinner('获取模块信息...');
      const ticket = yield def.auth.ticket({'app': 'def'});
      const locals = yield def.store.locals();
      mods = yield mods.map(m => def.store.diff(m.name, {'ticket': ticket}));
      mods = mods
        .filter(m => m.rel == 'gt')
        .filter(m => locals[m.name] && !locals[m.name].debug || !locals[m.name]);
      spinner.stop();

      if (mods.length == 0) {
        return def.log.info('[store]', '没有需要更新的模块');
      }

      console.log(def.ui.table(mods.map(m => {
        return {
          'name': m.name,
          'local': m.local,
          'remote': m.remote
        }
      })));

      if (!opts.noConfirm) {
        let confirmed = yield def.ui.confirm('确认执行更新？');
        if (!confirmed) {
          return def.log.info('[store]', '取消更新');
        }
      }

      let start = Date.now();
      def.log.info('[store]', '开始更新，请等待...');
      yield mods.map(m => def.store.update(m));
      def.log.info('[store]', '更新完成，耗时：' + (Date.now() - start) / 1000 + 's');
    });

  def.cli
    .command('remove <names...>', '删除本地安装的 DEF 模块')
    .alias('rm')
    .group('store')
    .action(function* (names) {
      const mods = names || [];

      def.log.info('[store]', '准备删除模块：' + mods.join(' '));
      const confirmed = yield def.ui.confirm('确认执行删除？');
      if (!confirmed) {
        return def.log.info('[store]', '取消删除');
      }

      const start = Date.now();
      def.log.info('[store]', '开始删除，请等待...');
      yield mods.map(m => def.store.remove(m));
      def.log.info('[store]', '删除完成，耗时：' + (Date.now() - start) / 1000 + 's');
    });

  def.cli
    .command('list', '列出本地安装的 DEF 模块')
    .alias('l')
    .argument('[type]', '列出的模块类型', ['plugin', 'kit', 'builder', 'generator'])
    .group('store')
    .action(function* (type) {
      let mods = yield def.store.list(type || '');
      if (mods.length == 0) {
        return def.log.info('[store]', '没有找到模块...');
      }

      const locals = yield def.store.locals();
      let groupedMods = {};
      mods.forEach(m => {
        if (!groupedMods[m.type]) {
          groupedMods[m.type] = [];
        }

        let mod = {
          'name': m.name,
          'local_version': m.version,
          'latest_version': 'missing',
          'beta_version': 'missing'
        };
        if (locals[m.name]) {
          if (locals[m.name].debug) {
            mod.dev_link = '-> ' + locals[m.name].realpath;
          }
          if (locals[m.name].version) {
            mod.latest_version = locals[m.name].version;
          }
          if (locals[m.name].next_version) {
            mod.beta_version = locals[m.name].next_version;
          }
        }

        groupedMods[m.type].push(mod);
      });

      let data = [], count = Object.keys(groupedMods).length;
      for (let k in groupedMods) {
        data = data.concat(groupedMods[k]);
        if (count > 1) {
          data = data.concat({
            'name': '',
            'local_version': '',
            'latest_version': '',
            'beta_version': '',
            'dev_link': ''
          });
        }
      }

      console.log();
      console.log(def.ui.table(data));
    });

};