'use strict';

const path = require('path');
const argv = process.argv.slice(2);
console.log('argv=', argv);

// argv= storeDir; server; api; tokenPath;
const storeFile = path.join(argv[0], '.store.json');
const lockFile = path.join(argv[0], '.store.lock');
const server = argv[1];
const api = argv[2];
const tokenPath = argv[3];

const co = require('co');
const io = require('./../io');
const fs = require('./../fs');
const auth = new (require('./../auth'))({
  'server': server,
  'tokenPath': tokenPath,
  'timeout': 60 * 1000
});

const TYPE_MAP = {
  '1': 'plugin',
  '2': 'kit',
  '3': 'generator',
  '4': 'builder'
};

const MAX_LOCKED_TIME = 10 * 60 * 1000; //10min


process.on('SIGINT', clear);
process.on('SIGTERM', clear);
process.on('exit', (code) => {
  console.log('exit', code);
  /*
   * code=
   *  0- success
   *  1- locked
   *  2- fatal
   */
  if (code != 1) {
    try {
      fs.unlinkSync(lockFile);
    } catch(e) {}
  }
});

function clear() {
  try {
    fs.unlinkSync(lockFile);
  } catch(e) {}
  process.exit(0);
}

co(function* () {
  let locked = false;
  try {
    locked = yield fs.existsAsync(lockFile);

    // check the lock time
    if (locked) {
      const now = Date.now();
      let lockedTime = yield fs.readFileAsync(lockFile, 'utf8');
      lockedTime = parseInt(lockedTime || now, 10);
      const lockedInterval = now - lockedTime;
      console.log(
        'lockedTime: %d; now: %d; interval: %d; max: %d',
        lockedTime, now, lockedInterval, MAX_LOCKED_TIME
      );

      if (lockedInterval > MAX_LOCKED_TIME) {
        console.log('reset to unlock');
        locked = false;
      }
    }
  } catch(e) {
    console.error('check lock error', e);
  }

  if (locked) {
    console.log('is locked');
    return 1;
  }

  console.log('lock');
  yield fs.writeFileAsync(lockFile, String(Date.now()), 'utf8');

  let localStore = {};
  try {
    localStore = yield fs.readJsonAsync(storeFile);
  } catch(e) {
    console.log('read error', e);
  }

  let ticket = '';

  // 已登录，则必须强制获取到ticket
  const token = yield auth.login({'silent': true});
  if (token) {
    ticket = yield auth.ticket({'app': 'def', 'silent': true});
  }

  let remoteMods = [], res;
  try {
    res = yield io.get(server + api, {'ticket': ticket}, {'timeout': 60 * 1000});
    if (res.body && res.body.data && res.body.data.mods) {
      remoteMods = res.body.data.mods || [];
    }
  } catch(e) {
    console.log('get mods error', e);
  }

  if (remoteMods.length > 0) {
    localStore.mods = localStore.mods || {};

    const newMods = {};
    remoteMods.forEach(mod => {
      if (localStore.mods[mod.name]) {
        newMods[mod.name] = Object.assign({}, localStore.mods[mod.name], {
          'type': TYPE_MAP[String(mod.type)],
          'description': mod.description,
          'version': mod.version,
          'next_version': mod.next_version,
          'is_next': mod.is_next,
          'changelog': mod.changelog,
          'enabled': mod.enabled,
          'showed': mod.showed
        });
      } else {
        newMods[mod.name] = {
          'type': TYPE_MAP[String(mod.type)],
          'name': mod.name,
          'description': mod.description,
          'version': mod.version,
          'next_version': mod.next_version,
          'is_next': mod.is_next,
          'changelog': mod.changelog,
          'enabled': mod.enabled,
          'showed': mod.showed
        };
      }
    });

    for (let k in localStore.mods) {
      // debug状态模块也需要加进去
      if (!newMods[k] && localStore.mods[k].debug) {
        newMods[k] = Object.assign({}, localStore.mods[k]);
      }
    }

    localStore.updated_at = Date.now();
    localStore.mods = newMods;
  }

  try {
    console.log('write store');
    yield fs.writeJsonAsync(storeFile, localStore);
  } catch(e) {
    console.log('write error', e);
  }

  return 0;
})
.then((c) => {
  console.log('store ok');
  process.exit(c);
})
.catch(e => {
  console.log('store error', e, e.stack);
  process.exit(2);
});