'use strict';

const path = require('path');
const debug = require('debug')('def:core');


module.exports = class Def {

  /*
   * opts: 
   * {
   *   'home',
   *   'env',
   *   'registry',
   *   'silent',
   *   'disableLog',
   *   'disableCheckUpdate',
   *   'store': {}
   * }
   */
  constructor(opts) {
    debug('constructor', opts);

    this._home = opts.home;
    this._pkg = require('../package.json');
    this._version = this._pkg.version;
    this._cliVersion = '';

    this._env = opts.env || 'prod';
    this._server = {
      'prod': 'https://def.taobao.com',
      'prepub': 'http://pre.def.taobao.com',
      'dev': 'http://deflite.daily.taobao.net',
      'local': 'http://local.daily.taobao.net:6001' // need host
    }[this._env];

    this._registry = opts.registry;

    this._verbose = false;

    this._fs = null;
    this._io = null;
    this._daemon = null;

    this._auth = null;
    this._store = null;
    this._ui = null;
    this._log = null;
    this._config = null;

    this._repo = null;
    this._builder = null;
    this._serve = null;
    this._yo = null;
    this._deploy = null;

    this._runtime = null;
    this._Promise = null;
    this._util = null;
    this._parser = null;

    this._kit = {};
    this._kitCommands = null;
    this._kits = null;

    this._plugin = {};
    this._plugins = {};

    this._parsed = null;

    this._disableTrack = process.env.DEF_TRACK == 'off';
    this._disableErrorlog = (process.env.DEF_ERROR_LOG == 'off' || process.env.DEF_ERRORLOG == 'off');
    this._enableDaemonlog = process.env.DEF_DAEMON_LOG == 'on';
    this._disableNotify = process.env.DEF_NOTIFY == 'off';
    this._disalbeCheckUpdate = process.env.DEF_CHECK_UPDATE == 'off';

    // silent mode, disable track/errorlog/notify
    if (opts.silent) {
      this._disableTrack = true;
      this._disableErrorlog = true;
      this._enableDaemonlog = false;
      this._disableNotify = true;
    }
    // disable log
    this._disableLog = false;
    if (opts.disableLog) {
      this._disableLog = true;
    }
    // disable check update
    if (opts.disableCheckUpdate) {
      this._disalbeCheckUpdate = true;
    }

    // custom store config
    this._storeConfig = opts.store || {};

    // def log level
    process.env.DEF_LOG_LEVEL = 'info';
  }

  /*
   * opts: {'clipkg', 'nodeVersion', 'argv', 'cwd'}
   */
  start(opts) {
    const def = this;
    const onFatal = def.onFatal.bind(def);

    opts = opts || {};
    debug('start', opts);

    process.on('uncaughtException', onFatal);
    process.on('unhandledRejection', onFatal);

    return require('co')(function* () {
      if (!def._disalbeCheckUpdate) {
        // 检查模块更新
        def._daemonCheck();
      }

      yield def.init(opts);
      yield def.run(opts);
    }).catch(onFatal);
  }

  /*
   * opts: {'clipkg', 'nodeVersion'}
   */
  * init(opts) {
    debug('init', opts);

    this._createCli(opts || {});
    yield this._init();
    yield this._registerCommands();
  }

  /*
   * opts: {'argv', 'cwd'}
   */
  * run(opts) {
    opts = opts || {};
    debug('run', opts);

    // set the correct argv back
    const rawArgv = opts.argv || process.argv;
    process.argv = rawArgv.slice();
    const argv = rawArgv.slice(2);
    debug('cli arg= %o', argv);

    // ch cwd if need
    opts.cwd = opts.cwd || process.cwd();
    if (process.cwd() !== opts.cwd) {
      process.chdir(opts.cwd);
      debug('chdir= %s', opts.cwd);
    };

    // pre parse to use plugin/kit.
    if (!this.plugin.key && !this.kit.key) {
      const parsed = yield this.cli.parse(argv, {'use': 'parser'});
      debug('cli pre parse', parsed);
      if (parsed.cmd) {
        this.log.verbose('[core]', 'cli pre use');
        try {
          yield this._preUse(parsed);
        } catch(e) {
          this.log.verbose('[core]', 'cli pre use error');
          return this.error(e);
        }
      }
    }

    // add the exit event
    this._onExit();

    // start cmd
    this._startTime = Date.now();
    let res = null;
    try {
      this.log.verbose('[core]', 'cli execute start');
      res = yield this.cli.parse(argv);
    } catch(e) {
      this.log.verbose('[core]', 'cli execute error');
      this.fatal(e);
    }
    if (res && res.$error) {
      this.log.verbose('[core]', 'cli execute $error= %s', res.$error);
      if ('invalid command' == res.$error) {
        this._endTime = Date.now();
        return yield this._suggestPlugin(res.cmd);
      }
    }

    this._endTime = Date.now();
    this.log.verbose('[core]', 'cli execute end. using: %ds', (this._endTime - this._startTime) / 1000);
    this.track({'time_used': this._endTime - this._startTime});
  }

  _createCli(opts) {
    this._nodeVersion = this._getNodeVersion(opts.nodeVersion);

    this._cliVersion = 'unknown';
    if (opts.clipkg && opts.clipkg.version) {
      this._cliVersion = opts.clipkg.version;
    }

    this.cli = require('./cli')(
      {
        'name': 'def',
        'description': 'DEF 客户端 for ISV',
        'version': '' +
          '     Cli: ' + (this._cliVersion == 'unknown' ? 'unknown' : 'v' + this._cliVersion) + '\n' +
          '    Core: v' + this._version + '\n' +
          '    Node: ' + this._nodeVersion + '\n' +
          'Registry: ' + this._registry,
        'lang': {
          'usage': '用法：',
          'arguments': '参数：',
          'options': '选项：',
          'commands': '命令：',

          'help': '显示帮助信息',
          'version': '显示版本信息',

          'error': '错误：',

          'invalidArgument': '非法的命令参数',
          'missingRequiredArgument': '缺少命令必须参数',

          'unknownOption': '不能识别的选项',
          'invalidOptionArgument': '非法的选项参数',

          'showHelp': '请查看帮助信息：'
        }
      }
    );
    this.cli.option('-V, --verbose', '显示详细日志信息');
    this.cli
      .on('parsed', this._onParsed.bind(this))
      .on('command.start', this._onCommandStart.bind(this))
  }

  * _init() {
    yield this.fs.ensureDirAsync(path.join(this._home, 'def_modules'));
    yield this.fs.ensureDirAsync(path.join(this._home, '.daemon'));
  }

  * _registerCommands() {
    require('./commands/auth')(this);
    require('./commands/store')(this);
    this._kitCommands = require('./commands/kit')(this);

    this._plugins = {};
    const plugins = yield this.store.list('plugin');
    plugins.forEach(p => this.registerPlugin(p));
  }

  * _preUse(parsed) {
    const info = this._getCmdInfo(parsed.cmd);
    this.log.verbose('[core]', 'preUse cmd= %j', info);

    if ('kit' == info.group) {
      let configJson = null;
      let kitType = 'unknown';
      let kitsType;

      const ctx = this;
      this._kits = yield this.store.fetch();
      this._kits.forEach(k => {
        ctx._kitCommands.init
          .command(k.type, k.desc)
          .action(function* (opts) {
            yield ctx.runKit('init', [k.type], opts, this, arguments);
          });
      });
      kitsType = this._kits.map(k => k.type);

      if (info.name == 'init') {
        this.log.verbose('[core]', 'guess type from arg');

        // try to get type from args. like: `def init xxx`
        kitType = parsed.argv._[0];
        if (kitsType.indexOf(kitType) != -1) {
          yield this.useKit(kitType);
        }
      } else {
        this.log.verbose('[core]', 'guess type from abc.json');

        try {
          configJson = this.lookupConfigJson();
        } catch (e) {
          throw new Error('项目 `abc.json` 文件解析失败: ' + e.message);
        }
        if (!configJson) {
          throw new Error('项目 `abc.json` 文件未找到，请检查执行目录是否正确');
        }

        kitType = configJson.kit || configJson.type || 'unknown';
        // FIXME: 映射 abs -> light
        if (kitType == 'abs') {
          kitType = 'light';
        }

        if (kitsType.indexOf(kitType) == -1) {
          yield this.useKit('default', {'type': kitType});
        } else {
          yield this.useKit(kitType);
        }
      }
    }

    if ('plugin' == info.group) {
      yield this.usePlugin(info.name);
    }
  }

  * _suggestPlugin(cmd) {
    this.log.verbose('[core]', 'suggest plugin %s', cmd);

    let info = yield this.store.diff('def-plugin-' + cmd);
    if (!info.error && info.rel == 'missing') {
      let confirm = yield this.ui.confirm('此命令可以通过安装插件 `' + info.long + '` 进行扩展，是否安装');
      if (!confirm) {
        this.log.info('[core]', '操作取消');
      } else {
        yield this.store.install(info);
      }
    } else {
      this.log.error('[core]', '无效命令 `%s`', cmd);
    }
  }

  _onExit() {
    const onExit = function() {
      debug('onExit');
      // 尚未打点，提前退出了
      if (!this._endTime && this._parsed) {
        this._endTime = Date.now();
        try {
          this.track({'time_used': this._endTime - this._startTime});
        } catch(e) {
          debug('onExit track error', e);
        }
      }
    }.bind(this);

    require('on-exit')(onExit);
  }

  _onParsed(data) {
    this._parsed = data.parsed;
    this.verbose = data.parsed.argv.verbose;
    debug('onParsed', data.parsed);
  }
  _onCommandStart(data) {
    this._cmdInfo = this._getCmdInfo(data.cmd);
    let realCmd = this._cmdInfo.name;
    if (this._cmdInfo.sub) {
      realCmd += ' ' + this._cmdInfo.sub;
    }
    this.log.verbose('[core]', 'cmd `%s` start with= %j', realCmd, data.args);
  }

  _getCmdInfo(cmd) {
    let cmdName = '';
    let cmdGroup = '';
    let cmdSub = '';
    if (cmd) {
      cmdName = cmd.name();
      cmdGroup = cmd.group();
      if (cmd._parent && cmd._parent.name() != 'def') {
        cmdName = cmd._parent.name();
        cmdGroup = cmd._parent.group();
        cmdSub = cmd.name();
      }
    }
    return {
      'name': cmdName,
      'group': cmdGroup,
      'sub': cmdSub
    };
  }

  registerPlugin(plugin) {
    debug('registerPlugin', plugin);

    if (!plugin) {
      return null;
    }

    const regNew = /^def-plugin-/;

    let pluginObj;
    let key = '';

    // 优先注册 def-plugin-xxx
    if (regNew.test(plugin.name)) {
      key = plugin.name.replace(regNew, '');
      pluginObj = _register.call(this);
    };

    return pluginObj;

    function _register() {
      const cmd = this.cli
        .command(key)
        .description(plugin.description)
        .version(plugin.version)
        .group('plugin')
        .action(function() { this.cmd.showHelp(); });

      const pluginObj = {
        'key': key,
        'name': plugin.name,
        'cmd': cmd,
        'published': {}
      };
      this._plugins[key] = pluginObj;

      return pluginObj;
    }
  }

  /*
   * key:
   * opts: {'pkg', 'fn', 'mockRuntime'}
   */
  * useKit(key, opts) {
    opts = opts || {};

    let name = 'def-kit-' + key;
    let defaultKit = require('./defaultKit')(this);
    let kitCmds = Object.keys(defaultKit.spec);

    let kitSpec;

    if (key == 'default') {
      if (opts.type !== 'unknown') {
        this.log.info('kit: ' + key + (opts.type ? (' (' + opts.type + ')') : ''));
      };

      this._kit = {
        '__type': opts.type,
        'key': opts.type,
        'name': opts.type
      };

      kitSpec = {};
    } else {
      let pkg = opts.pkg;
      if (!pkg) {
        pkg = yield this.checkUpdate(name, 'kit');
      }
      this._kit = {
        '__type': key,
        'key': key,
        'name': pkg.name,
        'version': pkg.version
      };

      // register init cmd.
      const ctx = this;
      this._kitCommands.init
        .command(key, pkg.description)
        .action(function* (opts) {
          yield ctx.runKit('init', [key], opts, this, arguments);
        });

      let fn = opts.fn;
      if (!fn) {
        fn = this.store.require(name);
      }
      let runtime = this.runtime.create({
        'key': key,
        'type': 'kit',
        'kit': this._kit
      });
      if (typeof opts.mockRuntime == 'function') {
        runtime = opts.mockRuntime(runtime);
      }
      kitSpec = fn(runtime);
    }

    const kitCommands = this._kitCommands;
    const kit = this._kit;

    // normalize hooks.
    kit.hooks = kitSpec.hooks || {};
    if (typeof kitSpec.before == 'function') {
      kit.hooks.before = kit.hooks.before || kitSpec.before;
    }
    if (typeof kitSpec.after == 'function') {
      kit.hooks.after = kit.hooks.after || kitSpec.after;
    }

    // register cmd.
    kitCmds.forEach(cmd => {
      let cmdSpec = kitSpec[cmd] || defaultKit.spec[cmd];
      let kitCmd = kitCommands[cmd];

      // normalize fn format.
      if (typeof cmdSpec == 'function') {
        kit[cmd] = {};
        kit[cmd].handler = cmdSpec;

        cmdSpec = kit[cmd];
      } else {
        kit[cmd] = cmdSpec;
      }

      if (cmdSpec.description) {
        kitCmd.description(cmdSpec.description);
      }

      cmdSpec.options = Object.assign(
        {},
        cmdSpec.options || {},
        cmdSpec.__no_default_options__ ? {} : defaultKit.spec[cmd].options
      );
      applyOptions(cmdSpec, kitCmd, cmdSpec.options);

      // custom add
      if (cmd == 'add') {
        let args = cmdSpec['arguments'];
        if (args) {
          let arg;
          if (arg = args.type) {
            kitCmd.argument('[type]', arg.description || '类型', arg.choices);
          }
          if (arg = args.names) {
            kitCmd.argument('[names...]', arg.description);
          }
        }
      }

      if (cmd == 'publish' && !cmdSpec.__force__) {
        defaultKit.registerPublishCmd(kitCmd);
      }
    });

    this.log.verbose('[core]', 'used kit');
    debug('useKit', kit);

    function applyOptions(commandSpec, command, options) {
      commandSpec.canChooseOptions = [];

      for (let name in options) {
        let opt = options[name];

        let flag = '--' + name;
        if (opt.alias) {
          flag = '-' + opt.alias + ', ' + flag;
        }
        let desc = opt.description || '';

        let config = {
          'type': opt.type,
          'choices': opt.choices,
          'default': opt.default,
          'hide': Boolean(opt.hide)
        };
        if (config.type != 'boolean') {
          if (-1 != ['string', 'number', 'array'].indexOf(config.type)) {
            flag += ' [' + name + (config.type == 'array' ? '...' : '') + ']';
          } else {
            if (null != config.default && 'boolean' != typeof config.default) {
              config.type = typeof config.default;
              if (-1 == ['string', 'number', 'array'].indexOf(config.type)) {
                config.type = 'string';
              }

              flag += ' [' + name + (config.type == 'array' ? '...' : '') + ']';
            }
          }
        }

        // chosed opts
        if (opt.canChoose && !config.hide) {
          commandSpec.canChooseOptions.push({
            'name': opt.description,
            'short': name,
            'value': name
          });
        }

        debug('applyOptions: %s; %s; %s; %j', command.name(), flag, desc, config);
        command.option(flag, desc, config);
      }
    }
  }

  /*
   * key:
   * opts: {'pkg', 'fn', 'mockRuntime'}
   */
  * usePlugin(key, opts) {
    opts = opts || {};

    let plugin = this._plugins[key];
    if (!plugin) {
      let info = opts.pkg;
      if (!info) {
        try {
          info = yield this.store.info('def-plugin-' + key);
        } catch (e) {}
        this.log.verbose('[core]', 'not found, check `%s` from store %j', key, info);
      }
      if (info) {
        plugin = this.registerPlugin({
          'name': info.name,
          'description': info.description,
          'version': info.version
        });
      }
    }

    let pkg = opts.pkg;
    if (!pkg) {
      pkg = yield this.checkUpdate(plugin.name, 'plugin');
    }
    plugin.version = pkg.version;

    let fn = opts.fn;
    if (!fn) {
      fn = this.store.require(plugin.name);
    }
    let runtime = this.runtime.create({
      'key': key,
      'type': 'plugin',
      'plugin': plugin
    });
    if (typeof opts.mockRuntime == 'function') {
      runtime = opts.mockRuntime(runtime);
    }
    fn(runtime);

    // FIXME: 兼容 ci 的“黑魔法”
    if (runtime.cli.__proto__.unknownOption) {
      runtime.cli.allowUnknownOption();
    }

    this.log.verbose('[core]', 'used plugin');
    debug('usePlugin', plugin);

    this._plugin = plugin;

    return plugin;
  }

  * checkUpdate(mod, type) {
    debug('check update: %s', mod);

    const spinner = this.ui.spinner('检查更新 (' + mod + ')...');
    const info = yield this.store.check(mod);
    spinner.stop();

    debug('mod info: %o', info);
    if (info.rel == 'invalid') {
      throw new Error(info.error);
    }

    if (!info.debug && !this._disalbeCheckUpdate) {
      let yes;
      let tip;
      if (info.rel == 'missing') {
        tip = '本地缺失, 开始安装 ' + info.long;
        if (info.remote == info.raw_next_version) {
          tip += '(灰度)'
        }
        tip += '，请等待...';

        this.log.info('[core]', tip);
        yield this.store.install(info);
      } else if (info.rel == 'gt') {
        tip = '模块 ' + mod + ' 有可用的新版本，' + info.local + ' -> ' + info.remote;
        if (info.remote == info.raw_next_version) {
          tip += '(灰度)';
        }
        this.log.info('[core]', tip);
        if (info.changelog && (info.remote != info.raw_next_version)) {
          console.log(this.ui.marked(info.changelog));
        }
        yes = yield this.ui.confirm('是否更新');

        if (yes) {
          this.log.info('[core]', '开始更新 ' + mod + '@' + info.remote + '，请等待...');
          yield this.store.update(info);
        }
      } else if (info.rel == 'lt') {
        this.log.info('[core]', '模块 ' + mod + ' 本地版本过高, 需降级后使用，' + info.remote + ' <- ' + info.local);
        yes = yield this.ui.confirm('是否降级');
        if (yes) {
          this.log.info('[core]', '开始降级 ' + mod + '@' + info.remote + '，请等待...');
          yield this.store.install(Object.assign({}, info, {'rel': 'missing'}));
        } else {
          this.log.error('[core]', '本地模块版本非法，退出执行');
          return process.exit(1);
        }
      }
    }

    let pkg = this.store.pkg(mod);
    if (!pkg) {
      throw new Error('can not require mod `' + mod + '`');
    }

    let spec = type + ': ' + pkg.name + '@' + pkg.version;
    if (!info.debug) {
      if (info.is_next || pkg.version == info.raw_next_version) {
        spec += ' (灰度)';
      }
    } else {
      spec += ' ' + this.ui.colors.blue('[调试中->' + info.realpath + ']');
      spec += ' ' + this.ui.colors.red('版本比对已被跳过');
    }
    this.log.info('[core]', spec);

    // debug模块禁用track&errorlog
    if (info.debug) {
      this._disableTrack = true;
      this._disableErrorlog = true;
    }

    return pkg;
  }

  onFatal(e, p) {
    debug('onFatal', e, p);
    this.fatal(e, p);
    process.exit(1);
  }

  * runKit(cmd, args, opts, actionCtx, actionArgs) {
    debug('runKit', cmd, args, opts, actionCtx, actionArgs);

    const kit = this.kit;

    let next = true;
    if (kit.hooks && typeof kit.hooks['before'] == 'function') {
      this.log.verbose('[core]', 'exec kit before hook');
      next = yield kit.hooks['before'](cmd, args, opts);
    }

    if (true == next || null == next) {
      let kitCmd = kit[cmd];
      if (kitCmd) {
        let choosedOpts = kitCmd.canChooseOptions || [];
        if (choosedOpts.length > 0) {
          let showChoice = true;
          let allOpts = Object.keys(opts).filter(o => o != '--' && o != '$');
          let choiceOpts = [];
          choosedOpts.forEach(opt => choiceOpts.push(opt.value));

          for (let i = 0, len = allOpts.length; i < len; i += 1) {
            let opt = allOpts[i];
            if (choiceOpts.indexOf(opt) != -1 && opts[opt] || 
              opt != 'verbose' && opts[opt]) {
              showChoice = false;
              break;
            }
          }
          if (showChoice) {
            let t = yield this.ui.list('请选择：', choosedOpts);
            opts[t] = true;
          }
        }

        if (cmd == 'dev' && !kit.dev.__no_reflect__) {
          // check port in use
          let inUse = yield checkPortInUse.call(this, opts.port);
          if (inUse) {
            this.log.error('[dev]', '启动 dev 服务失败，端口 `%d` 已经被占用。', opts.port);

            const yes = yield this.ui.confirm('是否使用其他端口启动 dev 服务？');
            if (!yes) {
              return this.log.info('[dev]', '你可以使用命令 def dev -p [port] 指定另外的端口。');
            }

            let port = yield this.ui.input('输入新端口号：', opts.port + 1111);
            port = parseInt(port, 10);

            inUse = yield checkPortInUse.call(this, port);
            if (inUse) {
              return this.log.info('[dev]', '该端口已经被占用，不可用。');
            }

            opts.port = port;
            this.log.verbose('[dev]', 'use the new port `%s`', port);
          }
        }

        if (kitCmd.action || kitCmd.handler) {
          this.log.verbose('[core]', 'exec kit action|handler');
          if (kitCmd.action) {
            yield kitCmd.action.apply(actionCtx, actionArgs);
          } else if (kitCmd.handler) {
            yield kitCmd.handler(args, opts);
          }
        }
      }
    }

    if (kit.hooks && typeof kit.hooks['after'] == 'function') {
      this.log.verbose('[core]', 'exec kit after hook');
      yield kit.hooks['after'](cmd, args, opts);
    }

    function* checkPortInUse(port) {
      const portUsed = require('tcp-port-used');

      let inUse = false;
      try {
        inUse = yield portUsed.check(port);
      } catch(e) {
        this.log.verbose('[dev]', 'check port error', e);
      }

      return inUse;
    }
  }

  get auth() {
    debug('get auth');
    if (this._auth) {
      debug('use cache');
      return this._auth;
    }
    this._auth = new (require('./auth'))({
      'server': this._server,
      'tokenPath': path.join(this._home, 'ACCESS_TOKEN')
    });
    debug('use require');
    return this._auth;
  }

  get store() {
    debug('get store');
    if (this._store) {
      debug('use cache');
      return this._store;
    }
    this._store = new (require('./store'))(Object.assign({
      'registry': this._registry,
      'storeDir': path.join(this._home, 'def_modules'),
      'infoApi': {
        'def': this._server + '/api'
      },
      'auth': this.auth
    }, this._storeConfig));
    debug('use require');
    return this._store;
  }

  get ui() {
    debug('get ui');
    if (this._ui) {
      debug('use cache');
      return this._ui;
    }
    this._ui = require('./ui');
    debug('use require');
    return this._ui;
  }

  get fs() {
    debug('get fs');
    if (this._fs) {
      debug('use cache');
      return this._fs;
    }
    this._fs = require('./fs');
    debug('use require');
    return this._fs;
  }

  get io() {
    debug('get io');
    if (this._io) {
      debug('use cache');
      return this._io;
    }
    this._io = require('./io');
    debug('use require');
    return this._io;
  }

  get log() {
    debug('get log');
    if (this._log) {
      debug('use cache');
      return this._log;
    }
    this._log = require('./log');
    debug('use require');
    return this._log;
  }

  get config() {
    debug('get config');
    if (this._config) {
      debug('use cache');
      return this._config;
    }
    this._config = require('./config');
    this._config.init(path.join(this._home, 'conf.json'));
    debug('use require');
    return this._config;
  }

  get runtime() {
    debug('get runtime');
    if (this._runtime) {
      debug('use cache');
      return this._runtime;
    }
    this._runtime = require('./runtime')(this);
    debug('use require');
    return this._runtime;
  }

  get verbose() {
    return this._verbose;
  }
  set verbose(v) {
    debug('set verbose %s', v);
    this._verbose = v;
    if (v) {
      this.log.level = 'verbose';
    } else {
      this.log.level = this._disableLog ? 'silent' : 'info';
    }
    process.env.DEF_LOG_LEVEL = this.log.level;
  }

  get daemon() {
    debug('get daemon');
    if (this._daemon) {
      debug('use cache');
      return this._daemon;
    }
    this._daemon = require('daemon').daemon;
    debug('use require');
    return this._daemon;
  }

  get repo() {
    debug('get repo');
    if (this._repo) {
      debug('use cache');
      return this._repo;
    }
    this._repo = require('./extensions/repo/')(this);
    debug('use require');
    return this._repo;
  }
  get serve() {
    debug('get serve');
    if (this._serve) {
      debug('use cache');
      return this._serve;
    }
    this._serve = require('./extensions/serve/')(this);
    debug('use require');
    return this._serve;
  }
  get builder() {
    debug('get builder');
    if (this._builder) {
      debug('use cache');
      return this._builder;
    }
    this._builder = require('./extensions/builder/')(this);
    debug('use require');
    return this._builder;
  }
  get yo() {
    debug('get yo');
    if (this._yo) {
      debug('use cache');
      return this._yo;
    }
    this._yo = require('./extensions/yo/')(this);
    debug('use require');
    return this._yo;
  }
  get deploy() {
    debug('get deploy');
    if (this._deploy) {
      debug('use cache');
      return this._deploy;
    }
    this._deploy = require('./extensions/deploy/')(this);
    debug('use require');
    return this._deploy;
  }

  get Promise() {
    debug('get Promise');
    if (this._promise) {
      debug('use cache');
      return this._promise;
    }
    this._promise = require('bluebird');
    debug('use require');
    return this._promise;
  }
  get util() {
    debug('get util');
    if (this._util) {
      debug('use cache');
      return this._util;
    }
    this._util = {
      'open': this.Promise.promisify(require('open')),
      'localip': this.Promise.promisify(require('node-localip'))
    };
    debug('use require');
    return this._util;
  }
  get parser() {
    debug('get parser');
    if (this._parser) {
      debug('use cache');
      return this._parser;
    }
    this._parser = require('yargs-parser');
    debug('use require');
    return this._parser;
  }

  get kits() {
    if (this._kits) {
      return this._kits;
    }
    return [];
  }

  get kit() {
    if (this._kit) {
      return this._kit;
    }
    return {};
  }
  get plugin() {
    if (this._plugin) {
      return this._plugin;
    }
    return {};
  }

  fatal(e, p) {
    debug('fatal', e, p);

    const reg = /^Cannot[ ]+find[ ]+module[ ]+'(.+)'$/;
    let m;
    if (m = (e.message || '').match(reg)) {
      let dep = String(m[1]);
      if (this._pkg.dependencies[dep]) {
        let where = path.join(__dirname, '../node_modules', dep);
        if (dep == 'cha-cli' || dep == 'grunt-cli') {
          where = path.join(where, 'bin', dep.replace('-cli', ''));
        }
        try {
          require.resolve(where);
        } catch(e) {
          e.message = 'missing dependency `' + dep + '`, try `def repair` to fix';
          return this.error(e);
        }
      }
    }
    this.error(e, p);

    // collect fatal error.
    let form = {};

    form['code'] = e.code || '';
    form['message'] = e.message || '';

    form['stack_info'] = e.stack || '';
    form['stack_info'] = (this.log.record || []).map(item => {
      const arr = [];
      if (item.level) arr.push(item.level);
      if (item.prefix) arr.push(item.prefix);
      // strip ansi
      arr.push(item.message.replace(/[\u001b\u009b][[()#;?]*(?:[0-9]{1,4}(?:;[0-9]{0,4})*)?[0-9A-ORZcf-nqry=><]/g, ''));
      return arr.join(' ');
    }).join('\n');

    if (!this._parsed) {
      this.log.verbose('[core]', 'there is no parsed data');
      return;
    }

    form = this._collectForm(form);

    // dump to file always
    try {
      this.fs.writeFileSync(path.join(this._home, 'def-debug.log'), form['stack_info'], 'utf8');
    } catch (e) {}

    if (this._disableErrorlog) {
      return;
    }

    this.log.verbose('[core]', 'do errorlog');
    this._doReport('/api/errorlog', form);
  }

  track(form) {
    debug('track');

    form = this._collectForm(form || {});

    if (this._disableTrack) {
      return;
    }

    this.log.verbose('[core]', 'do track');
    this._doReport('/api/track', form);
  }

  error(e, p) {
    if (!e) {
      this.log.error('[core]', 'unknown fatal error');
      return;
    }
    if (p) {
      this.log.verbose('[core]', 'unhandledRejection promise: ', p);
    }

    this.log.error('[core]', e.message || 'unknown error');

    const errorStack = e.stack || 'unknown error stack';
    if (this.verbose) {
      console.error(errorStack);
    } else {
      this.log.record.push({'level': '', 'prefix': '', 'message': errorStack});
    }
  }

  _doReport(api, form) {
    const argv = [
      this._server,
      api,
      path.join(this._home, 'ACCESS_TOKEN'),
      JSON.stringify(form)
    ];
    this._createDaemon(path.join(__dirname, 'daemons', 'report.js'), argv, 'report');
  }

  _collectForm(form) {
    form = form || {};

    const info = this._getCmdInfo(this._parsed.cmd);

    form['os'] = process.platform || 'unknown';
    form['node_ver'] = process.version || 'unknown';
    form['def_ver'] = this._cliVersion + '_' + this._version;

    form['cmd']  = 'def ' + this._parsed.raw.join(' ');
    form['rel_cmd'] = form['cmd'];

    form['callee'] = 'unknown';
    if (info.group == 'kit') {
      form['callee'] = this.kit.__type;
      form['mod_ver'] = this.kit.version || '';
    } else {
      form['callee'] = info.name || 'unknown';
      form['mod_ver'] = this.plugin.version || '';
    }
    form['rel_mod'] = form['callee'];

    // 1-plugin; 2-kit; 4-basic(auth,store); 3-others
    const typeMap = {
      'plugin': 1,
      'kit': 2,
      'auth': 4,
      'store': 4
    };
    form['mod_type'] = info.group ? (typeMap[info.group] || 3) : 3;

    if (info.name == 'init' || info.name == 'publish') {
      form['action'] = info.name;
      form['args'] = info.sub ? info.sub : this._parsed.argv._.join(' ');
    } else {
      form['action'] = info.sub ? info.sub : info.name;
      form['args'] = this._parsed.argv._.join(' ');
    }

    debug('collect form= %j', form);
    return form;
  }

  _daemonCheck() {
    debug('daemon check');
    const argv = [
      path.join(this._home, 'def_modules'),
      this._server, '/api/mods/all',
      path.join(this._home, 'ACCESS_TOKEN')
    ];
    this._createDaemon(path.join(__dirname, 'daemons', 'check.js'), argv, 'check');
  }

  lookupConfigJson(cwd) {
    debug('lookupConfigJSon');
    const abcFile = lookupConfigFile.call(this, cwd);
    debug('configFile: %s', abcFile);
    if (!abcFile) {
      return null;
    }

    let configJson = this.fs.readFileSync(abcFile, 'utf8');
    configJson = require('json-parse-helpfulerror').parse(configJson);
    debug('configJson: %j', configJson);
    return configJson;

    function lookupConfigFile(cwd) {
      cwd = cwd || process.cwd();

      let abcFile = path.join(cwd, 'abc.json');
      if (this.fs.existsSync(abcFile)) {
        return abcFile;
      };
      return '';
    }
  }

  _createDaemon(file, argv, key) {
    let p = null;
    try {
      let logConfig = {};
      if (this._enableDaemonlog) {
        logConfig = {
          'stdout': this.fs.openSync(path.join(this._home, '.daemon', key + '_out.log'), 'w'),
          'stderr': this.fs.openSync(path.join(this._home, '.daemon', key + '_error.log'), 'w')
        };
      }
      debug('daemon %s= %j', key, argv);
      p = this.daemon(file, argv, logConfig);
    } catch(e) {
      debug('daemon error', key, e);
    }
    return p;
  }

  _getNodeVersion(nodeVersion) {
    let str = '';
    if (!nodeVersion) {
      str += 'unknown';
    } else {
      nodeVersion = nodeVersion || {};
      str += (nodeVersion.bundle ? nodeVersion.bundle : 'unknown') + '(bundle); ';
      str += (nodeVersion.system ? nodeVersion.system : 'unknown') + '(system); ';
      str += 'using ' + (nodeVersion.use ? nodeVersion.use : 'unknown');
    }
    return str;
  }

}
