'use strict';


module.exports = function(def) {
  const fs = require('fs');
  const path = require('path');
  const co = require('co');

  const serve = {};

  serve._cpServer = null;

  serve.cmdOpts = {};
  serve.start = co.wrap(start.bind(serve));
  serve.fork = co.wrap(fork.bind(serve));

  return serve;

  /**
   *
   * @param opts
   *   - cwd{String}
   *
   *   - type{String}
   *   - builderReflect{String}
   *   - customServer{Boolean}
   *
   *   - livereload{Boolean}
   *   - watchDir{String}
   *   - livereloadServerImpl{Function}
   *   - customLivereload{Boolean}
   *
   *   - disableCombo{Boolean}
   *   - disableIndex{Boolean}
   *
   *   - server{Function}
   *
   *   - * 其他均会透传
   */
  function* start(opts) {
    opts = normalizeOpts(opts);
    def.log.verbose('[dev]', 'start with: ', opts);

    yield checkPort(opts.port);

    const reflectJs = yield getReflectJs(opts);

    if (opts.customServer) {
      def.log.verbose('[dev]', 'use custom server');
      return require(reflectJs);
    }
    return yield require('./app').start(require(reflectJs), opts);
  }

  function* fork(opts, forkOpts) {
    opts = normalizeOpts(opts);
    forkOpts = Object.assign({'cwd': opts.cwd}, forkOpts || {});

    def.log.verbose('[dev]', 'fork with: ', opts, forkOpts);

    if (!serve._cpServer) {
      const cp = require('child_process');
      const stringify = require('json-stringify-safe');

      yield checkPort(opts.port);

      const reflectJs = yield getReflectJs(opts);

      serve._cpServer = {
        'child_process': null,
        '_fn': function() {},
        '_events': function() {
          const ctx = this;
          this.child_process.on('message', function(message) {
            def.log.verbose('[dev]', 'forked cp on message', message);

            if (message.__error) {
              const e = new Error(message.__error.message);
              e.stack = message.__error.stack;
              ctx._fn('error', e);
            } else {
              ctx._fn('message', message);
            }
          });
          this.child_process.on('exit', function(code) {
            def.log.verbose('[dev]', 'forked cp exit', code);

            ctx._fn('exit', code);
          });
          this.child_process.on('error', function(err) {
            def.log.verbose('[dev]', 'forked cp on error', err);

            ctx._fn('error', err);
          });
        },
        '_fork': function(startOpts) {
          if (this.child_process) {
            return;
          }
          
          this.child_process = cp.fork(
            path.join(__dirname, 'server.js'),
            [
              '--app=' + path.join(__dirname, 'app.js'),
              '--reflect=' + reflectJs, 
              '--opts=' + stringify(Object.assign({}, opts, startOpts))
            ],
            forkOpts
          );
        },

        'send': function(message) {
          message = Object.assign(message || {}, {'__from': 'sender'});

          def.log.verbose('[dev]', 'send `%j` to forked cp', message);

          if (this.child_process) {
            this.child_process.send(message);
          }
        },

        'start': function(startOpts) {
          def.log.verbose('[dev]', 'fork start');

          if (!this.child_process) {
            this._fork(startOpts);
            this._events();
          }

          this.child_process.send({'__from': 'main', '__action': 'start'});

          return this;
        },
        'stop': function() {
          def.log.verbose('[dev]', 'fork stop');

          // if (this.child_process) {
          //   this.child_process.send({'__from': 'main', '__action': 'stop'});
          // }
          this.child_process.kill();
          this.child_process = null;

          return this;
        },
        'restart': function() {
          def.log.verbose('[dev]', 'fork restart');

          // if (this.child_process) {
          //   this.child_process.send({'__from': 'main', '__action': 'restart'});
          // }
          this._fork();
          this.start();
          
          return this;
        },

        'on': function(fn) {
          this._fn = fn;
        }
      };

      if (!forkOpts.manualStart) {
        serve._cpServer._fork();
        serve._cpServer.start();
        serve._cpServer._events();
      }

    }

    // 存在子进程已经关闭的情况，用 try catch 进行容错
    process.on('exit', function(){
      try {
        serve._cpServer.stop();
      }catch(e){}
    });
    // catches ctrl+c event
    process.on('SIGINT', function(){
      try {
        serve._cpServer.stop();
      }catch(e){}
    });
    //catches uncaught exceptions
    process.on('uncaughtException', function(){
      try {
        serve._cpServer.stop();
      }catch(e){}
    });

    return serve._cpServer;
  }

  function* getReflectJs(opts) {
    let reg = /^(builder-[-_a-zA-Z]+)\/(reflect\.js|server\.js)$/;

    let builder = '', reflectJs = 'reflect.js';
    let configJson = def.lookupConfigJson() || {};

    if (opts.builderReflect) {
      let match = opts.builderReflect.match(reg);
      if (!match) {
        return def.log.error('[dev]', 'builderReflect `%s` 参数非法', opts.builderReflect)
      }
      builder = match[1];
      reflectJs = match[2];
    }
    if (!builder) {
      def.log.verbose('[dev]', 'lookup builder from `abc.json`');
      if (configJson.options && configJson.options.builderReflect && configJson.builder) {
        builder = configJson.builder;
      }
    }

    if (builder) {
      yield def.checkUpdate(builder, 'builder');

      def.log.info('[dev]', '使用 `%s/%s` 启动服务器', builder, reflectJs);
      return def.store.resolve(builder + '/' + reflectJs);
    }

    let type = opts.type || '';
    if (!type) {
      def.log.verbose('[dev]', 'lookup type from `abc.json`');
      if (configJson.type) {
        type = configJson.type;
      }
    }
    let configJs = path.join(__dirname, 'config', type) + '.js';
    if (!fs.existsSync(configJs)) {
      def.log.info('[dev]', '无法加载 `%s` 配置, 使用 `default` 配置', type);
      type = 'default';
      configJs = path.join(__dirname, 'config', type) + '.js';
    }
    def.log.info('[dev]', '使用 `%s` 配置启动服务器', type);
    return configJs;
  }

  function* checkPort(port) {
    const inUse = yield require('tcp-port-used').check(port);
    if (inUse) {
      throw new Error('port `' + port + '` is used');
    }
    return port;
  }

  function normalizeOpts(opts) {
    opts = Object.assign(serve.cmdOpts, opts || {});

    opts = Object.assign({
      'port': 3333,
      'debug': false,
      'disableIndex': 'index' in opts ? !opts.index : false,
      'livereload': 'noLivereload' in opts ? false : true
    }, opts);

    if (opts.lrPort == null) {
      opts.lrPort = opts.port + 32369;
    }

    opts.cwd = opts.cwd || process.cwd();

    return opts;
  }

};
