'use strict';

const debug = require('debug')('def:installer');

const assert = require('assert');
const path = require('path');

const Npmii = require('npminstall');
const onExit = require('on-exit');

const fs = require('./fs');
const Util = require('./util');

const platform = process.platform;

/**
 *
 * @param registry
 * @returns {Object}
 */
module.exports = function(registry) {
  const api = {};
  api.registry = registry;
  api.npmi = npmi;
  api.npmii = npmii;
  return api;
};

/**
 *
 * @param m
 * @param opts
 */
function* npmi(m, opts) {
  opts = opts || {};
  debug('npmi', opts);

  assert(opts.storeDir != '', '[Installer] missing opts.storeDir');
  assert(opts.cacheDir != '', '[Installer] missing opts.cacheDir');

  const config = {
    'cmd': 'npm',
    'root': path.join(opts.storeDir, m.name),
    'name': m.name,
    'version': m.version
  };

  yield fs.removeAsync(config.root);
  yield Npmi(config);
}

/**
 *
 * @param m
 * @param opts
 */
function* npmii(m, opts) {
  opts = opts || {};
  debug('npminstall opts', opts);

  assert(opts.storeDir != '', '[Installer] missing opts.storeDir');
  assert(opts.cacheDir != '', '[Installer] missing opts.cacheDir');

  const lockFile = path.join(opts.storeDir, '.', m.name, '.lock');
  const locked = yield fs.existsAsync(lockFile);
  if ((platform != 'win32') && locked && !opts.force) {
    throw new Error('锁释放异常，无法完成操作，请执行 `def i ' + m.name + ' -f` 进行强制操作');
  }

  // the succeed mark
  let succeed = false;

  // lock the module
  if (platform != 'win32') {
    yield fs.ensureFileAsync(lockFile);
    onExit(clear);
  }

  const utils = require('npminstall/lib/utils');

  const env = {'npm_config_registry': this.registry};
  env['npm_node_execpath'] = env.NODE = process.env.NODE || process.execPath;
  /* FIXME: other env
   * - npm_config_argv= 
   * - npm_execpath= require.main.filename
   * - npm_rootpath= 
   */

  const binaryMirros = yield utils.getBinaryMirrors(this.registry);
  for (const key in binaryMirros.ENVS) {
    env[key] = binaryMirros.ENVS[key];
  }

  // set node-gyp env for windows.
  if (process.platform === 'win32') {
    let python = Util.getPython();
    let msvsVersion = Util.getMsvsVersion();
    if (python) {
      env['npm_config_python'] = python;
    }
    if (msvsVersion) {
      env['npm_config_msvs_version'] = msvsVersion;
    }
  }

  process.env.NO_PROXY = '*';

  const config = {
    'production': true,
    'cacheDir': '',
    'env': env,
    'registry': this.registry,
    'binaryMirrors': binaryMirros,
    'timeout': opts.timeout || 10 * 60 * 1000,
    'strictSSL': Util.getStrictSSL(),
    'ignoreScripts': false,
    'root': path.join(opts.storeDir, m.name),
    'pkgs': [m]
  };
  if (opts.flatten) {
    config.root = path.dirname(opts.storeDir);
  }
  debug('npmii config', config);

  if (opts.flatten) {
    try {
      yield fs.removeAsync(path.join(opts.storeDir, m.name));
    } catch(e) {}
  } else {
    const realDir = path.join(config.root, 'node_modules');
    const exists = yield fs.existsAsync(realDir);
    if (exists) {
      try {
        if (platform == 'win32') {
          yield fs.removeAsync(realDir);
          if (platform == 'win32') {
            // FIX: window 删除文件后延迟下
            yield sleep(250);
          }
        } else {
          const trashFile = path.join(opts.storeDir, '.' + m.name.replace('/', '_') + '_' + m.version + '_' + Date.now());
          yield fs.moveAsync(realDir, trashFile);
          try {
            require('daemon').daemon(path.join(__dirname, 'daemons', 'trash.js'), [trashFile]);
          } catch(e) {}
        }
      } catch(e) {
        clear();
        throw e;
      }
    }
  }

  try {
    yield Npmii(config);

    succeed = true;
    clear();
  } catch(e) {
    clear();
    throw e;
  }

  function clear() {
    if (platform == 'win32') return;

    try {
      fs.unlinkSync(lockFile);
      if (!succeed) {
        fs.unlinkSync(path.join(opts.storeDir, m.name, 'node_modules', m.name, 'package.json'));
      }
    } catch(e) {}
  }

  function sleep(ms) {
    return function(cb) {
      setTimeout(cb, ms);
    };
  }
}


/**
 *
 * @param opts
 * @returns {bluebird}
 * @constructor
 */
function Npmi(opts) {
  debug('npm i opts', opts);
  return new Promise(function(resolve, reject) {
    let error = '';
    let pkgFile = path.join(opts.root, 'package.json');
    let pkgData = {
      'name': opts.name,
      'version': opts.version,
      'private': true
    };

    fs.ensureFileSync(pkgFile);
    fs.writeJsonSync(pkgFile, pkgData);

    const cmd = Util.spawnCommand(opts.cmd || 'npm',
      ['i'].concat([opts.name + '@' + opts.version, '--loglevel=http']),
      {'cwd': opts.root, 'stdio': 'inherit'}
    );
    cmd.on('error', reject);
    cmd.on('exit', function(code) {
      if (code == 0) {
        return resolve();
      }
      return reject(new Error(code));
    });
  });
}