'use strict';

const debug = require('debug')('def:store');
const assert = require('assert');

const path = require('path');

const semver = require('semver');
const globby = require('globby');
const requireStack = require('require-stack');

const fs = require('./fs');
const io = require('./io');
const Util = require('./util');
const Installer = require('./installer');


module.exports = class Store {

  constructor(opts) {
    opts = opts || {};

    assert(opts.registry != '', '[Store] missing opts.registry.');
    assert(opts.storeDir != '', '[Store] missing opts.storeDir.');
    assert(opts.infoApi != null, '[Store] missing opts.infoApi.');
    assert(opts.auth != null, '[Store] missing opts.auth.');

    this.auth = opts.auth;
    this.infoApi = opts.infoApi; // {'def', 'builder'}
    this.registry = opts.registry;
    this.timeout = opts.timeout || 10 * 1000;

    this.storeDir = opts.storeDir;
    this.storeFile = path.join(this.storeDir, '.store.json');
    this.builderStoreDir = path.join(this.storeDir, '.builders');
    this.generatorStoreDir = path.join(this.storeDir, '.generators');

    // 拉平模式下均存储在一级目录
    this.flatten = opts.flatten;
    if (this.flatten) {
      this.builderStoreDir = this.storeDir;
      this.generatorStoreDir = this.storeDir;
    }

    this.installer = Installer(this.registry);
  }

  require(p) {
    let filepath = this._getPath(p);
    debug('require', p, filepath);

    return requireStack(filepath);
  }

  resolve(p) {
    let filepath = this._getPath(p);
    debug('resolve', p, filepath);

    if (fs.existsSync(filepath)) {
      return filepath;
    }
    return '';
  }

  pkg(mod) {
    let filepath = path.join(
      this._realRoot(mod), Util.generatePath(mod, 'package.json', this.flatten)
    );
    debug('pkg', mod, filepath);

    let data = null;
    try {
      data = fs.readJsonSync(filepath, {'throw': false})
    } catch(e) {}
    return data;
  }

  * install(diffInfo, opts) {
    opts = opts || {};

    if (diffInfo.error) {
      throw new Error(diffInfo.error);
    }

    const info = this._normalize(diffInfo, opts);
    debug('install', diffInfo, opts, info);

    // 本地缺失
    if (diffInfo.rel == 'missing') {
      debug('install.missing');
      yield this.installer[info.method](info.mod, info.config);
      return true;
    }

    // 强制覆盖
    if (opts.force) {
      debug('install.force');
      info.config = Object.assign({}, info.config, {'force': true});
      yield this.installer[info.method](info.mod, info.config);
    }
    return true;
  }

  * update(diffInfo, opts) {
    opts = opts || {};

    if (diffInfo.error) {
      throw new Error(diffInfo.error);
    }

    const info = this._normalize(diffInfo, opts);
    debug('update', diffInfo, opts, info);

    // 远程版本更新
    if (diffInfo.rel == 'gt') {
      yield this.installer[info.method](info.mod, info.config);
      return true;
    }
    return true;
  }

  * updateOrInstall(diffInfo, opts) {
    opts = opts || {};

    if (diffInfo.error) {
      throw new Error(diffInfo.error);
    }

    const info = this._normalize(diffInfo, opts);
    debug('updateOrInstall', diffInfo, opts, info);

    if (diffInfo.rel == 'missing' || diffInfo.rel == 'gt') {
      yield this.installer[info.method](info.mod, info.config);
      return true;
    }
    if (opts.force) {
      info.config = Object.assign({}, info.config, {'force': true});
      yield this.installer[info.method](info.mod, info.config);
    }
    return true;
  }

  * remove(mod) {
    const p = path.join(this._realRoot(mod), mod);
    debug('remove', mod, p);
    yield fs.removeAsync(p);
  }

  * list(type) {
    type = type || '';

    let pats = [
      ['def-*', '!def-kit-*'],
      'def-kit-*',
      'builder-*',
      'generator-*'
    ];
    let dirs = [];

    if (type == 'kit') {
      dirs = yield globby(pats[1], {'cwd': this.storeDir});
    } else if (type == 'builder') {
      dirs = yield globby(pats[2], {'cwd': this.builderStoreDir});
    } else if (type == 'generator') {
      dirs = yield globby(pats[3], {'cwd': this.generatorStoreDir});
    } else if (type == 'plugin') {
      dirs = yield globby(pats[0], {'cwd': this.storeDir});
      dirs = _sortPlugins(dirs);
    } else {
      dirs = yield globby(pats[0], {'cwd': this.storeDir});
      dirs = _sortPlugins(dirs);

      dirs = dirs.concat(yield globby(pats[1], {'cwd': this.storeDir}));
      dirs = dirs.concat(yield globby(pats[2], {'cwd': this.builderStoreDir}));
      dirs = dirs.concat(yield globby(pats[3], {'cwd': this.generatorStoreDir}));
    }

    dirs = yield dirs.map(d => {
      const p = this.pkg(d);

      let o = {
        'name': d,
        'version': 'missing',
        'description': 'missing',
        'type': 'unknown'
      };
      if (p) {
        o.name = p.name;
        o.version = p.version;
        o.description = p.description;
        o.type = Util.type(o.name);
      }
      return o;
    });

    return dirs;

    function _sortPlugins(dirs) {
      const reg = /def-plugin/;
      return dirs.sort((a, b) => {
        const testa = reg.test(a);
        const testb = reg.test(b);
        if (testa && testb) {
          return 0;
        }
        if (testa) {
          return -1;
        }
        if (testb) {
          return 1;
        }
        return 0;
      });
    }
  }

  * diff(mod, query) {
    let info;
    let data = {
      'name': mod,
      'long': mod,
      'local': this._local(mod),
      'remote': '',
      'rel': '',
      'is_next': false,
      'changelog': '',
      'latest': '',
      'next': '',
      'error': ''
    };

    let status;
    try {
      info = yield this.info(mod, query);
      data.remote = info.version;
      data.is_next = info.is_next;
      data.changelog = info.changelog;
      data.latest = info.raw_version;
      data.next = info.raw_next_version;
    } catch(e) {
      status = e.status;
      if (status == 404) {
        data.error = 'DEF Store 中不存在模块 `' + mod + '`';
      } else if (status == 401) {
        data.error = 'DEF Store 中尚未启用模块 `' + mod + '`';
      } else {
        data.error = e.message;
      }
    }

    if (data.remote) {
      data.long += '@' + data.remote;
    }

    if (!data.local) {
      data.rel = 'missing';
    } else {
      if (!data.remote) {
        data.rel = 'exits';
        if (status == 404 || status == 401) {
          data.rel = 'invalid';
        }
      } else {
        data.rel = this._rel(data.remote, data.local);
      }
    }

    debug('diff', data);

    return data;
  }

  * check(mod) {
    const mods = yield this.locals();

    let data;

    let m = mods[mod];
    if (!m) {
      let query = {};
      query.ticket = yield this._ticket();
      data = yield this.diff(mod, query);

      debug('check from network', data);
    } else {
      data = {};

      data.name = m.name;
      data.local = this._local(data.name);
      data.remote = m.version;
      data.is_next = Boolean(m.is_next);
      if (data.is_next) {
        data.remote = m.next_version || m.version;
      }
      data.changelog = String(m.changelog) || '';

      data.raw_version = m.version;
      data.raw_next_version = m.next_version || '';

      data.long = data.name;
      if (data.remote) {
        data.long += '@' + data.remote;
      }

      data.error = '';
      data.debug = Boolean(m.debug);
      data.realpath = m.realpath;

      if (!data.local) {
        data.rel = 'missing';
      } else {
        if (!data.remote) {
          data.rel = 'debug';
          if (!data.debug) {
            data.rel = 'invalid';
          }
        } else {
          data.rel = this._rel(data.remote, data.local);
        }
      }

      debug('check from local', data);
    }

    return data;
  }

  * info(mod, query) {
    query = query || {};

    let info = Util.info(mod);
    let url = this.infoApi.def + '/mod/' + encodeURIComponent(mod);

    if (!query.ticket) {
      query.ticket = yield this._ticket();
    }

    debug('info.get', url, query);

    let res = yield io.get(url, query, {
      'timeout': this.timeout
    });
    if (res.body.error) {
      const err = new Error(res.body.message);
      err.status = res.body.status;
      throw err;
    }

    let data = {};
    if (info.type == 'builder') {
      data = res.body.data;
    } else {
      data = res.body.data.mod;
    }

    let o = {};
    o.is_next = data.is_next;
    o.name = data.name || mod;
    o.version = data.version;
    if (o.is_next) {
      o.version = data.next_version || data.version;
    }
    o.description = data.description || '';
    o.changelog = data.changelog || '';
    
    o.raw_version = data.version;
    o.raw_next_version = data.next_version || '';

    debug('info.done', data, o);

    return o;
  }

  * fetch(opts) {
    opts = opts || {};

    let kits = [];

    const locals = yield this.locals();
    debug('local mods in store.fetch', locals);

    Object.keys(locals).forEach(k => {
      const info = Util.info(k);
      if (info.type == 'kit') {
        const kit = locals[k];
        kits.push({
          'desc': kit.description,
          'name': info.name,
          'type': info.name,
          'showed': kit.showed
        });
      }
    });

    debug('local kits in store.fetch', kits);

    if (kits.length == 0) {
      try {
        const res = yield io.get(this.infoApi.def + '/kit', {}, {
          'timeout': opts.timeout || this.timeout
        });
        if (res.body.data && res.body.data.kits && res.body.data.kits.length) {
          kits = res.body.data.kits.map(function(kit) {
            const o = {};
            o.desc = kit.description;
            o.type = kit.name.replace('def-kit-', '');
            o.name = o.type;
            o.showed = kit.showed;
            return o;
          });
        }
      } catch(e) {}
    }

    return kits;
  }

  * locals() {
    let mods = {};
    try {
      const json = yield fs.readJsonAsync(this.storeFile);
      if (json && json.mods) {
        mods = json.mods;
      }
    } catch(e) {}

    return mods;
  }

  join() {
    return Array.prototype.slice.call(arguments).join('/');
  }

  * _ticket() {
    debug('_ticket()');
    if (!this.ticket) {
      debug('get ticket from network');
      try {
        this.ticket = yield this.auth.ticket({
          'app': 'def',
          'timeout': this.timeout,
          'silent': true
        });
      } catch(e) {}
    }
    return this.ticket || '';
  }

  _local(mod) {
    const p = this.pkg(mod);
    if (p) {
      return p.version;
    }
    return '';
  }

  _rel(a, b) {
    let r = 'eq';
    if (semver.gt(a, b)) {
      r = 'gt';
    } else if (semver.lt(a, b)) {
      r = 'lt';
    }
    return r;
  }

  _getPath(p) {
    const parts = p.split('/');

    let name = parts.shift();
    let suffix = parts.join('/');

    return path.join(
      this._realRoot(name), Util.generatePath(name, suffix, this.flatten)
    );
  }

  _realRoot(name) {
    if (Util.isGenerator(name)) {
      return this.generatorStoreDir;
    } else if (Util.isBuilder(name)) {
      return this.builderStoreDir;
    } else {
      return this.storeDir;
    }
  }

  _normalize(diffInfo, opts) {
    const mod = {
      'name': diffInfo.name,
      'version': diffInfo.remote
    };
    const config = {
      'storeDir': this._realRoot(diffInfo.name),
      'force': opts.force,
      'flatten': this.flatten
    };
    const method = opts.useNpmi ? 'npmi' : 'npmii';

    return {
      'mod': mod,
      'config': config,
      'method': method
    };
  }

}